from pybullet_envs.scene_abstract import SingleRobotEmptyScene
from pybullet_envs.env_bases import MJCFBaseBulletEnv
import numpy as np
import pybullet
import gym

import real_robots

from .robot import Kuka
import sys, os

def DefaultRewardFunc(observation):
    return 0

class Goal:
    def __init__(self, initial_state=None, final_state=None, retina=None,
                retina_before = None, challenge = None):
        self.initial_state = initial_state
        self.final_state = final_state
        self.retina = retina
        self.retina_before = retina_before
        self.challenge = challenge

class REALRobotEnv(MJCFBaseBulletEnv):
    """ Create a REALCompetion environment inheriting by gym.env

    """

    intrinsic_timesteps = int(1e7)
    extrinsic_timesteps = int(2e3)
    
    def __init__(self, render=False):

        self.robot = Kuka()
        MJCFBaseBulletEnv.__init__(self, self.robot, render)
        
        self._cam_dist = 1.2
        self._cam_yaw = 30
        self._cam_roll = 0
        self._cam_pitch = -30
        self._render_width = 320
        self._render_height = 240
        self._cam_pos = [0,0,.4]
        self.setCamera()
        self.eyes = {}

        self.reward_func = DefaultRewardFunc
        
        self.robot.used_objects = ["table", "tomato", "mustard", "cube"]
        self.set_eye("eye")

        self.goal = Goal(retina=self.observation_space.spaces[
            self.robot.ObsSpaces.GOAL].sample()*0)
        
        # Set default goals dataset path
        # 
        # The goals dataset is basically a list of real_robots.envs.env.Goal
        # objects which are stored using : 
        # 
        # np.savez_compressed(
        #               "path.npy.npz",
        #                list_of_goals)
        #
        self.goals_dataset_path = os.path.join( 
                                    real_robots.getPackageDataPath(),
                                    "goals_dataset.npy.npz")
        self.goals = None
        self.goal_idx = -1
   
    def setCamera(self):
        ''' Initialize environment camera
        '''
        self.envCamera = EnvCamera(
                distance=self._cam_dist, 
                yaw=self._cam_yaw,
                pitch=self._cam_pitch, 
                roll=self._cam_roll, 
                pos=self._cam_pos,
                width=self._render_width,
                height=self._render_height)
    
    def set_eye(self, name, eye_pos=[0.01, 0, 1.2], target_pos=[0, 0, 0]):
        ''' Initialize an eye camera
        @name the label of the created eye camera
        '''
        cam = EyeCamera(eye_pos, target_pos)
        self.eyes[name] = cam

    def set_goals_dataset_path(self, path):
        assert os.path.exists(path), "Non existent path {}".format(path)
        self.goals_dataset_path = path
        self.goals = list(np.load(
                self.goals_dataset_path, allow_pickle=True).items())[0][1]
        self.goal_idx = 0
        
    def set_goal(self):
        if self.goals is None:
            self.goals = list(np.load(
                    self.goals_dataset_path, allow_pickle=True).items())[0][1]
            self.goal_idx = 0
        self.goal = self.goals[self.goal_idx]
        
        for obj in self.goal.initial_state.keys():
            self.robot.object_bodies[obj].reset_position(self.goal.initial_state[obj][:3])

        self.goal_idx += 1

    def extrinsicFormula(self, p_goal, p, a_goal, a, w = 1):
        pos_dist = np.linalg.norm(p_goal-p)
        pos_const = - np.log(0.25) / 0.05 # Score goes down to 0.25 within 5cm 
        pos_value = np.exp(- pos_const * pos_dist)

        orient_dist = min(np.linalg.norm(a_goal-a),np.linalg.norm(a_goal+a))
        orient_const = - np.log(0.25) / 0.30 # Score goes down to 0.25 within 0.3
        orient_value = np.exp(- orient_const * orient_dist)

        value = w * pos_value + (1-w) * orient_value
        return value

    def evaluateGoal(self):
        initial_state = self.goal.initial_state
        final_state = self.goal.final_state
        current_state = self.robot.object_bodies

        score = 0
        n_obj = len(final_state.keys())
        for obj in final_state.keys():
            p = np.array(current_state[obj].get_position())
            p_goal = np.array(final_state[obj][:3])
            a = np.array(current_state[obj].get_pose()[3:])
            a_goal = np.array(final_state[obj][3:])

            if self.goal.challenge == '3D':
                w = 0.75
            else:
                w = 1

            objScore = self.extrinsicFormula(p_goal, p, a_goal, a, w) 
            # print("Object: {} Score: {:.4f}".format(obj,objScore))
            score += objScore / n_obj

        # print("Goal score: {:.4f}".format(score))
        return self.goal.challenge, score

        
    def create_single_player_scene(self, bullet_client):
        return SingleRobotEmptyScene(bullet_client, gravity=9.81, 
                timestep=0.005, frame_skip=1)
    
    def reset(self):

        super(REALRobotEnv, self).reset()
        self._p.setGravity(0.,0.,-9.81)
        self.camera._p = self._p
        for name in self.eyes.keys():
           self.eyes[name]._p = self._p
        
        self._p.resetDebugVisualizerCamera(
                self._cam_dist, self._cam_yaw, 
                self._cam_pitch, self._cam_pos)

        self.timestep = 0
        
        return self.get_observation()

    def render(self, mode='human', close=False):
            if mode == "human":
                    self.isRender = True
            if mode != "rgb_array":
                    return np.array([])

            rgb_array = self.envCamera.render(self._p)
            return rgb_array

    def get_part_pos(self, name):
        #print(self.robot.parts.keys())
        return self.robot.parts[name].get_position()
    
    def get_obj_pos(self, name):
        return self.robot.object_bodies[name].get_position()

    def get_obj_pose(self, name):
        return self.robot.object_bodies[name].get_pose()

    def get_contacts(self):
        return self.robot.get_contacts()
    
    def get_retina(self):
        '''
        :return: the current rgb_array for the eye
        '''
        return self.eyes["eye"].render(self.robot.object_bodies["table"].get_position()) 
 
    def control_objects_limits(self):
        '''
        reset positions if an object goes out of the limits
        '''
        for obj in self.robot.used_objects:
            x, y, z = self.robot.object_bodies[obj].get_position()
            if not ( -0.2 < x < 0.2) or not ( -0.5 < y < 0.5) or z < 0.33:
                self.robot.object_bodies[obj].reset_position(
                        self.robot.object_poses[obj][:3])

    def get_observation(self):

        joints = self.robot.calc_state()
        sensors = self.robot.get_touch_sensors()
        retina = self.get_retina()
        
        observation = {
                Kuka.ObsSpaces.JOINT_POSITIONS: joints,
                Kuka.ObsSpaces.TOUCH_SENSORS: sensors,
                Kuka.ObsSpaces.RETINA: retina,
                Kuka.ObsSpaces.GOAL: self.goal.retina }

        return observation


    def step(self, action):
        assert(not self.scene.multiplayer)
        
        self.control_objects_limits()
        self.robot.apply_action(action)
        self.scene.global_step()  

        
        observation = self.get_observation()

        reward = self.reward_func(observation)       
        
        done = False
        if self.goal_idx < 0:
            if self.timestep >= self.intrinsic_timesteps:
                done = True
        else:
            if self.timestep >= self.extrinsic_timesteps:
                done = True

        info = {}
        
        self.timestep += 1

        return observation, reward, done, info

class REALRobotEnvSingleObj(REALRobotEnv):
    def __init__(self, render=False):
        super(REALRobotEnvSingleObj, self).__init__(render)
        self.robot.used_objects = ["table", "cube"]

class EnvCamera:

    def __init__(self, distance, yaw, pitch, roll, pos, 
            fov=80, width=320, height=240):
        
        self.dist = distance
        self.yaw = yaw
        self.pitch = pitch
        self.roll = roll
        self.pos = pos
        self.fov = fov
        self.render_width = width
        self.render_height = height

    def render(self, bullet_client = None):
        
        if bullet_client is None:
            bullet_client = self._p

        view_matrix = bullet_client.computeViewMatrixFromYawPitchRoll(
                cameraTargetPosition = self.pos,
                distance=self.dist,
                yaw=self.yaw,
                pitch=self.pitch,
                roll=self.roll,
                upAxisIndex=2)

        proj_matrix = bullet_client.computeProjectionMatrixFOV(
                fov=self.fov, aspect=float(self.render_width)/self.render_height,
                nearVal=0.1, farVal=100.0)

        (_, _, px, _, _) = bullet_client.getCameraImage(
                width=self.render_width, height=self.render_height,
                viewMatrix=view_matrix,
                projectionMatrix=proj_matrix,
                renderer=pybullet.ER_BULLET_HARDWARE_OPENGL
                )

        rgb_array = np.array(px).reshape(self.render_height, self.render_width, 4)
        rgb_array = rgb_array[:, :, :3]

        return rgb_array
     
class EyeCamera:

    def __init__(self, eyePosition, targetPosition,
            fov=80, width=320, height=240):
        
        self.eyePosition = eyePosition
        self.targetPosition = targetPosition
        self.upVector = [0, 0, 1]
        self.fov = fov
        self.render_width = width
        self.render_height = height
        self._p = None
        self.pitch_roll = False
    
    def render(self, *args, **kargs):
        if self.pitch_roll is True:
            return self.renderPitchRoll(*args, **kargs)
        else:
            return self.renderTarget(*args, **kargs)


    def renderTarget(self, targetPosition, bullet_client = None):
        
        if bullet_client is None:
            bullet_client = self._p

        self.targetPosition = targetPosition

        view_matrix = bullet_client.computeViewMatrix(
                cameraEyePosition = self.eyePosition,
                cameraTargetPosition = self.targetPosition,
                cameraUpVector=self.upVector)

        proj_matrix = bullet_client.computeProjectionMatrixFOV(
                fov=self.fov, aspect=float(self.render_width)/self.render_height,
                nearVal=0.1, farVal=100.0)

        (_, _, px, _, _) = bullet_client.getCameraImage(
                width=self.render_width, height=self.render_height,
                viewMatrix=view_matrix,
                projectionMatrix=proj_matrix,
                renderer=pybullet.ER_BULLET_HARDWARE_OPENGL
                )

        rgb_array = np.array(px).reshape(self.render_height, self.render_width, 4)
        rgb_array = rgb_array[:, :, :3]

        return rgb_array
            
    def renderPitchRoll(self, distance, roll, pitch, yaw, bullet_client = None):
        
        if bullet_client is None:
            bullet_client = self._p

        self.targetPosition = targetPosition

        view_matrix = bullet_client.computeViewMatrixFromYawPitchRoll(
                cameraTargetPosition = self.pos,
                distance=distance,
                yaw=yaw,
                pitch=pitch,
                roll=roll,
                upAxisIndex=2)

        proj_matrix = bullet_client.computeProjectionMatrixFOV(
                fov=self.fov, aspect=float(self.render_width)/self.render_height,
                nearVal=0.1, farVal=100.0)

        (_, _, px, _, _) = bullet_client.getCameraImage(
                width=self.render_width, height=self.render_height,
                viewMatrix=view_matrix,
                projectionMatrix=proj_matrix,
                renderer=pybullet.ER_BULLET_HARDWARE_OPENGL
                )

        rgb_array = np.array(px).reshape(self.render_height, self.render_width, 4)
        rgb_array = rgb_array[:, :, :3]

        return rgb_array
            





# -*- coding: utf-8 -*-

import gym
from .envs import Goal
import numpy as np

from tqdm.auto import tqdm

"""Local evaluation helper functions."""


def build_score_object(scores):
    score_string = ""
    total_score = 0
    challenges = ['2D','2.5D','3D']

    score_object = {}
    for key in challenges:
        if key in scores.keys():
            results = scores[key]
            formatted_results = ", ".join(["{:.4f}".format(r) for r in results])
            challenge_score = np.mean(results)
        else:
            results = []
            formatted_results = "None"
            challenge_score = 0
        total_score += challenge_score
        score_object["score_{}".format(key)] = challenge_score
    total_score /= len(challenges)

    score_object["score_total"] = total_score
    return  score_object


def evaluate(Controller, 
            intrinsic_timesteps=1e7,
            extrinsic_timesteps=2e3,
            extrinsic_trials=350,
            visualize=True,
            goals_dataset_path="./goals.npy.npz"):
    """
    A wrapper function to locally simulate the evaluation process
    as is done for all the submitted controllers.

    Parameters
    ----------
    Controller 
        An example controller which should expose a `step` function, for 
        the evaluator to compute the `action` given observation, reward
        and done info
    
    intrinsic_timesteps : int
        Maximum number of timesteps in the Intrinsic phase
    extrinsic_timesteps: int
        Maximum number of timesteps in the Extrinsic phase
    extrinsic_trials: int
        Total number of trials in the extrinsic phase
    visualize: bool
        Boolean flag which enables or disables the visualizer when 
        running the evaluation
    goals_dataset_path: str
        Path to a goals dataset
    """
    env = gym.make('REALRobot-v0')
    env.set_goals_dataset_path(goals_dataset_path)

    if visualize:
        env.render('human')
    
    controller = Controller(env.action_space)

    env.intrinsic_timesteps = intrinsic_timesteps #default = 1e7
    env.extrinsic_timesteps = extrinsic_timesteps #default = 2e3
    extrinsic_trials = 3

    ##########################################################
    ##########################################################
    # Helper functions
    ##########################################################
    ##########################################################
    scores = {}
    def add_scores(challenge, score):
        if challenge in scores.keys():
            scores[challenge] += [score]
        else:
            scores[challenge] = [score]

    ##########################################################
    ##########################################################    
    observation = env.reset()
    reward = 0
    done = False    
    intrinsic_phase_progress_bar = tqdm(
                        total=intrinsic_timesteps,
                        desc="Intrinsic Phase",
                        unit="steps ",
                        leave=False
                        )
    intrinsic_phase_progress_bar.write("######################################################")
    intrinsic_phase_progress_bar.write("# Intrinsic Phase Initiated")
    intrinsic_phase_progress_bar.write("######################################################")

    # intrinsic phase
    steps = 0
    while not done:
        # Call your controller to chose action 
        action = controller.step(observation, reward, done)
        # do action
        observation, reward, done, _ = env.step(action)
        steps += 1
        intrinsic_phase_progress_bar.update(1)
    intrinsic_phase_progress_bar.write("######################################################")
    intrinsic_phase_progress_bar.write("# Intrinsic Phase Complete")
    intrinsic_phase_progress_bar.write("######################################################")
    # extrinsic phase
    # tqdm.write("Starting extrinsic phase")

    extrinsic_phase_progress_bar = tqdm(
                                        total = extrinsic_trials,
                                        desc = "Extrinsic Phase Trials",
                                        unit = "trials ",
                                        leave=True
                                        )
    extrinsic_phase_progress_bar.write("######################################################")
    extrinsic_phase_progress_bar.write("# Extrinsic Phase Initiated")
    extrinsic_phase_progress_bar.write("######################################################")

    totalScore = 0
    for k in range(extrinsic_trials):
        observation = env.reset()
        reward = 0
        done = False
        extrinsic_phase_progress_bar.update(1)
        env.set_goal()

        while not done:
            action = controller.step(observation, reward, done)
            observation, reward, done, _ = env.step(action)

        add_scores(*env.evaluateGoal())
        extrinsic_phase_progress_bar.set_postfix(
                        build_score_object(
                                scores
                            )
                    )

    extrinsic_phase_progress_bar.write("######################################################")
    extrinsic_phase_progress_bar.write("# Extrinsic Phase Complete")
    extrinsic_phase_progress_bar.write("######################################################")
    extrinsic_phase_progress_bar.write(str(build_score_object(scores)))
    return build_score_object(scores)



from robot_bases import MJCFBasedRobot, URDFBasedRobot
import numpy as np
import pybullet_data
import os
import gym
from robot_bases import BodyPart


class Kuka(URDFBasedRobot):
    
    used_objects = [
            "table",
            "orange",
            "mustard",
            "cube",
            "tomato"]

    object_poses = {
            "table":   [ 0.00,  0.00,  0.000, 0.00, 0.00, 0.00],
            "orange":  [-0.10,  0.00,  0.460, 0.00, 0.00, 0.00],
            "mustard": [-0.05, -0.40,  0.480, 1.57, 3.14, 1.57],
            "cube":    [-0.10,  0.00,  0.470, 0.00, 0.00, 0.00], 
            "tomato":  [-0.10,  0.40,  0.485, 0.00, 0.00, 0.00]}

    num_joints = 9
    num_kuka_joints = 7
    num_gripper_joints = 2
    num_touch_sensors = 4
    eye_width = 320
    eye_height = 240
    
    class ObsSpaces: 
        JOINT_POSITIONS = "joint_positions"
        TOUCH_SENSORS = "touch_sensors"
        RETINA = "retina"
        GOAL = "goal"

    def __init__(self):

        self.robot_position = [-0.8, 0, 0]
        self.contact_threshold = 0.1

        self.action_dim = self.num_joints
        
        URDFBasedRobot.__init__(self, 
                'kuka_gripper_description/urdf/kuka_gripper.urdf', 
                'kuka0', action_dim=self.action_dim, obs_dim=1)
              
        self.observation_space = gym.spaces.Dict({
            self.ObsSpaces.JOINT_POSITIONS: gym.spaces.Box(
                -np.inf, np.inf, [self.num_joints], dtype = float),
            self.ObsSpaces.TOUCH_SENSORS: gym.spaces.Box(
                0, np.inf, [self.num_touch_sensors], dtype = float),
            self.ObsSpaces.RETINA: gym.spaces.Box(
                0, 255, [Kuka.eye_height, Kuka.eye_width, 3], dtype = float),
            self.ObsSpaces.GOAL: gym.spaces.Box(
                0, 255, [Kuka.eye_height, Kuka.eye_width, 3], dtype = float)})

        self.target = "orange"

        self.object_names = dict()
        self.object_bodies = dict()
        self.robot_parts = {}

 
    def reset(self, bullet_client):
        bullet_client.resetSimulation()
        super(Kuka, self).reset(bullet_client)
        return self.calc_state()
    
    def get_contacts(self, forces=False):    

        contact_dict = {}
        for part_name, part in self.parts.items():         
            contacts = []
            for contact in part.contact_list():
                if abs(contact[8]) < self.contact_threshold:
                    name = self.object_names[contact[2]] 
                    if not forces:
                        if part_name in contact_dict.keys():
                            contact_dict[part_name].append(name)
                        else:
                            contact_dict[part_name]= [name]  
                    else:
                        force = contact[9]
                        if part_name in contact_dict.keys():
                            contact_dict[part_name].append([name, force])
                        else:
                            contact_dict[part_name] = [(name, force)]  

        return contact_dict
    
    def get_touch_sensors(self): 
        
        sensors = np.zeros(4)
        contacts = self.get_contacts(forces=True)
        for i, skin in enumerate(["skin_00", "skin_01", "skin_10", "skin_11"]):
            if skin in contacts.keys():
                cnts = contacts[skin]   
                if len(cnts) > 0:
                    force = np.max([cnt[1] for cnt in cnts])
                    sensors[i] = force
                
        return sensors 

    def robot_specific_reset(self, bullet_client):

        self.robot_body.reset_position(self.robot_position)

        self.object_bodies["kuka"] = self.robot_body
        self.object_names[0] = "kuka"

        for obj_name in self.used_objects:
            pos = self.object_poses[obj_name]
            obj = get_object(bullet_client,
                    "kuka_gripper_description/urdf/{}.urdf".format(obj_name),
                    *pos)
            self.object_bodies[obj_name] = obj
            self.object_names.update({obj.bodies[0]: obj_name})
        
        for _,joint in self.jdict.items():
            joint.reset_current_position(0, 0)

        for name, part in self.parts.items():
            self.robot_parts.update({part.bodyPartIndex: name})

    def apply_action(self, a):
        assert (np.isfinite(a).all())
        assert(len(a) == self.num_joints)

        a[:-2] = np.maximum(-np.pi*0.5, np.minimum(np.pi*0.5, a[:-2]))
        a[-2:] = np.maximum( 0, np.minimum(np.pi*0.5, a[-2:]))
        a[-1] = np.maximum( 0, np.minimum(2*a[-2], a[-1]))
 
        for i,j in enumerate(a[:-2]):
            self.jdict["lbr_iiwa_joint_%d"%(i+1)].set_position(j)
        
        self.jdict["base_to_finger00_joint"].set_position(a[-2])
        self.jdict["base_to_finger10_joint"].set_position(a[-2])
        self.jdict["finger00_to_finger01_joint"].set_position(-a[-1])
        self.jdict["finger10_to_finger11_joint"].set_position(-a[-1])

    def calc_state(self):
        joints = []
        for i in range(self.num_kuka_joints):
            joints.append(self.jdict["lbr_iiwa_joint_%d"%(i+1)].get_position())
        joints.append(self.jdict["base_to_finger00_joint"].get_position())
        joints.append(-self.jdict["finger00_to_finger01_joint"].get_position())
        
        return joints 



 
def get_object(bullet_client, object_file, x, y, z, roll=0, pitch=0, yaw=0):

    position = [x, y, z]
    orientation = bullet_client.getQuaternionFromEuler([roll, pitch, yaw])
    fixed = True
    body = bullet_client.loadURDF(
            fileName=os.path.join(pybullet_data.getDataPath(), object_file),
            basePosition=position,
            baseOrientation=orientation,
            useFixedBase=False,
            flags=bullet_client.URDF_USE_INERTIA_FROM_FILE)
    part_name, _ = bullet_client.getBodyInfo(body)
    part_name = part_name.decode("utf8")
    bodies = [body]
    return BodyPart(bullet_client, part_name, bodies, 0, -1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import os

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

def get_all_data_files(directory='./real_robots/data/'):
    ret = []
    for dirpath, subdirs, files in os.walk(directory):
        for f in files:
            ret.append(os.path.join(dirpath, f))
    return ret


requirements = ['Click>=6.0', 'gym', 'pybullet', 'numpy', 'tqdm']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="S.P. Mohanty",
    author_email='mohanty@aicrowd.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Robots that learn to interact with the environment autonomously",
    entry_points={
        'console_scripts': [
            'real_robots=real_robots.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='real_robots',
    name='real_robots',
    packages=find_packages(include=['real_robots']),
    data_files=[('data', get_all_data_files())],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/spMohanty/real_robots',
    version='0.1.3',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""Top-level package for real-robots."""

__author__ = """S.P. Mohanty"""
__email__ = 'mohanty@aicrowd.com'
__version__ = '0.1.3'

import os
from gym.envs.registration import register

register(id='REALRobot-v0',
    entry_point='real_robots.envs:REALRobotEnv',
)

register(id='REALRobotSingleObj-v0',
    entry_point='real_robots.envs:REALRobotEnvSingleObj',
)

from .evaluate import evaluate


def getPackageDataPath():
    import real_robots
    return os.path.join(
                real_robots.__path__[0],
                "data"
            )

def copy_over_data_into_pybullet(force_copy=False):
    """
    If the package specific data has not already 
    been copied over into pybullet_data, then
    copy them over.
    """
    import pybullet_data

    pybullet_data_path = pybullet_data.getDataPath()
    is_data_absent = "kuka_gripper_description" not in os.listdir(pybullet_data_path)
    if force_copy or is_data_absent:
        import shutil
        source_data_path = os.path.join(
                                getPackageDataPath(),
                                "kuka_gripper_description")
        target_data_path = os.path.join(
                                pybullet_data_path,
                                "kuka_gripper_description")
        print(
            "[REALRobot] Copying over data into pybullet_data_path."
            "This is a one time operation.")
        shutil.copytree(source_data_path, target_data_path)
copy_over_data_into_pybullet()